@@ -53,55 +53,56 @@ module REXML
 		end
 
 		# @return the root of the document that this element belongs to.
-	 # If this element doesn't belong to a document, but does belong to
-	 # another Element, the parent's root will be returned, until the
-	 # earliest ancestor is found.
-	 def root
-		 parent.nil? ? self : parent.root
-	 end
+		# If this element doesn't belong to a document, but does belong to
+		# another Element, the parent's root will be returned, until the
+		# earliest ancestor is found.
+		def root
+			parent.nil? ? self : parent.root
+		end
 
-	 def whitespace
-		 @whitespace = (@context and @context[:respect_whitespace] and
-		 (@context[:respect_whitespace] == :all or
-		 @context[:respect_whitespace].include? expanded_name))
-		 @whitespace
-	 end
+		def whitespace
+			@whitespace = (@context and @context[:respect_whitespace] and
+				(@context[:respect_whitespace] == :all or
+				 @context[:respect_whitespace].include? expanded_name)
+			)
+			@whitespace
+		end
 
-	 def raw
-		 @raw = (@context and @context[:raw] and
-		 (@context[:raw] == :all or
-		 @context[:raw].include? expanded_name))
-		 @raw
-	 end
+		def raw
+			@raw = (@context and @context[:raw] and
+			(@context[:raw] == :all or
+			@context[:raw].include? expanded_name))
+			@raw
+		end
 
-	 once :whitespace, :raw
+		once :whitespace, :raw
 
-	 #################################################
-	 # Namespaces                                    #
-	 #################################################
+		#################################################
+		# Namespaces                                    #
+		#################################################
 
-	 # @return an Array of all defined namespaces
-	 def prefixes
-		 prefixes = []
-		 prefixes = parent.prefixes if parent
-		 prefixes += attributes.prefixes
-		 return prefixes
-	 end
+		# @return an Array of all defined namespaces
+		def prefixes
+			prefixes = []
+			prefixes = parent.prefixes if parent
+			prefixes += attributes.prefixes
+			return prefixes
+		end
 
-	 # @return the URI for a namespace, or the empty string if no such 
-	 # namespace is declared
-	 def namespace prefix=nil
-		 if prefix.nil?
-			 prefix = prefix()
-		 end
-		 prefix = "xmlns" if prefix == ""
-		 ns = attributes[ prefix ]
-		 ns = parent.namespace(prefix) if ns.nil? and parent
-		 return ns
-	 end
+		# @return the URI for a namespace, or the empty string if no such 
+		# namespace is declared
+		def namespace prefix=nil
+			if prefix.nil?
+				prefix = prefix()
+			end
+			prefix = "xmlns" if prefix == ""
+			ns = attributes[ prefix ]
+			ns = parent.namespace(prefix) if ns.nil? and parent
+			return ns
+		end
 
-	 # Adds a namespace to this element
-	 # @param prefix the namespace.  If the second argument is not supplied,
+		# Adds a namespace to this element
+		# @param prefix the namespace.  If the second argument is not supplied,
 		# this should be the URI of the default namespace.
 		# @param value (optional) the URI of the namespace.  If "", then unsets
 		# the default namespace
@@ -193,7 +194,7 @@ module REXML
 		# @param xpath any XPath by which to search for elements in the tree
 		def get_elements( xpath )
 			rv = []
-			each( xpath ) {|child| rv<<child }
+			@elements.each( xpath ) {|child| rv<<child }
 			rv
 		end
 
@@ -286,7 +287,7 @@ module REXML
 		# @return self.  NOTE that contrary to most REXML methods, the object
 		# added to the parent is not returned.
 		def add_text( text )
-			text = Text.new( text ) if text.kind_of? String
+			text = Text.new( text, whitespace(), raw() ) if text.kind_of? String
 			self << text unless text.nil?
 			return self
 		end
@@ -319,6 +320,18 @@ module REXML
 			end
 		end
 
+		# Add multiple attributes to this element.  EG:
+		# el.add_attributes( {"name1"=>"value1", "name2"=>"value2"} )
+		# el.add_attributes( [ ["name1","value1"], ["name2"=>"value2"] ] )
+		# @p hash is either a hash, or array of arrays
+		def add_attributes hash
+			if hash.kind_of? Hash
+				hash.each_pair {|key, value| @attributes[key] = value }
+			elsif hash.kind_of? Array
+				hash.each { |value| @attributes[ value[0] ] = value[1] }
+			end
+		end
+
 		# Removes an attribute
 		# @param key either an Attribute or a String.  In either case, the
 		# attribute is found by matching the attribute name to the argument,
@@ -354,43 +367,50 @@ module REXML
 		end
 
 		def Element.parse_stream source, listener
-			# Get the next tag
-			md = source.match(/^<([\w-]*:?[\w-]+)(.*?>)/um, true)
-			name = md[1]
-			rest = md[2]
-			attrs = rest.scan(Attribute::PATTERN)
-			attrs.each { |ar| ar[1,1] = [] }
-			listener.tag_start name, attrs
-			unless rest =~ /\/>$/u		# empty
-				while true
-					md = source.match(/\A(\s*[^>]*>)/um)
-					raise ParseException.new("malformed XML: bad content", source, 
+			begin
+				# Get the next tag
+				name, closed, attrs = base_parser(source)
+				attrs.each { |ar| ar[1,1] = [] }
+				listener.tag_start name, attrs
+				unless closed
+					while true
+						md = source.match(/\A(\s*[^>]*>)/um)
+						raise ParseException.new("malformed XML: bad content", source, 
 						self) unless md
-					line = md[1]
-					return if line.nil?
-					case line
-					when /\A<\w/um
-						Element.parse_stream source, listener
-					when /\A<\//um
-						break
-					when CData::START_RE
-						CData.parse_stream source, listener
-					when Comment::START_RE
-						Comment.parse_stream source, listener
-					when Instruction::START_RE
-						Instruction.parse_stream source, listener
-					else
-						Text.parse_stream source, listener
+						line = md[1]
+						return if line.nil?
+						case line
+						when /\A<\w/um
+							Element.parse_stream source, listener
+						when /\A<\//um
+							break
+						when CData::START_RE
+							CData.parse_stream source, listener
+						when Comment::START_RE
+							Comment.parse_stream source, listener
+						when Instruction::START_RE
+							Instruction.parse_stream source, listener
+						else
+							Text.parse_stream source, listener
+						end
 					end
-				end
-				if source.match(/^\s*<\/#{name}\s*>/um, true)
-					listener.tag_end name
-				else
-					raise ParseException.new( "Missing end tag for #{expanded_name}", 
+					if source.match(/^\s*<\/#{name}\s*>/um, true)
+						listener.tag_end name
+					else
+						raise ParseException.new( "Missing end tag for #{name}", 
 						source, self )
+					end
+				else
+					listener.tag_end name
 				end
-			else
-				listener.tag_end name
+			rescue ParseException
+				$!.source = source
+				$!.element = self
+				raise
+			rescue Exception
+				raise
+				old_ex = $!
+				raise ParseException.new("unidentified error", source, self, old_ex)
 			end
 		end
 
@@ -405,100 +425,95 @@ module REXML
 			}
 		end
 
-
-		def parse source
-			debug = false
-			begin
-				# Get the next tag
-				md = source.match(/^<((?:[:!\w_][-\.\w_]*:)?[-!\*\.\w_]*)\s*((((["']).*?\5)|[^\/'">]*)*?)(\/)?(>)/um, true)
-				raise ParseException.new("malformed XML: missing tag start", source,
-					self) unless md
-				puts "TAG: Matched \"#{md[0]}\", buffer is \"#{source.buffer}\"" if debug
-				puts "MATCH DATA = [#{md.to_a.join ', '}]" if debug
-				self.name = md[1]
-				attributes = md[2]
-				slash = md[6]
-				close = md[7]
-				if attributes
-					attrs = attributes.scan( Attribute::PATTERN )
+		# A helper method, to consolidate the common code between parse and
+		# parse_stream.
+		# @return an array with [tag_name_S, closed_B, attributes_A]
+		def Element.base_parser source
+			# Get the next tag
+			md = source.match(/^<((?:[:\w_][-\.\w_]*:)?[-!\*\.\w_]*)\s*((((["']).*?\5)|[^\/'">]*)*?)(\/)?>/um, true)
+			raise ParseException.new("malformed XML: missing tag start", source,
+				self) unless md
+				attrs = []
+				if md[2]
+					attrs = md[2].scan( Attribute::PATTERN )
 					raise ParseException.new( "error parsing attributes: [#{attrs.join ', '}], excess = \"#$'\"", source, self ) if $' and $'.strip.size > 0
-					#puts "ATTRIBUTES: Matched [#{attrs.join ', '}], buffer is #{source.buffer}" if debug
+				end
+				return [md[1], md[6], attrs]
+			end
+
+			def parse source
+				begin
+					# Get the next tag
+					self.name, closed, attrs = Element.base_parser(source)
 					attrs.each do |set|
 						@attributes.add( Attribute.new( set[0], set[2] ) )
 					end
+
+					unless closed
+						parse_children source
+						raise ParseException.new( "Missing end tag for #{expanded_name}",
+							source, self) unless source.match(/^\s*<\/#@expanded_name\s*>/um, 
+							true)
+						end
+					rescue ParseException
+						$!.source = source
+						$!.element = self
+						raise
+					rescue Exception
+						old_ex = $!
+						raise ParseException.new("unidentified error", source, self, old_ex)
+					end
 				end
 
-				puts "CLOSE TAG: Matched #{close}, slash is #{slash}" if debug
-				if close and not slash
-					parse_children source
-					raise ParseException.new( "Missing end tag for #{expanded_name}",
-						source, self) unless source.match(/^\s*<\/#@expanded_name\s*>/um, 
-						true)
-				else
-					raise ParseException.new( "Malformed tag for #{expanded_name}",
-						source, self) unless close
+				# A private helper method
+				def parse_children source
+					while true
+						md = source.match(/\A(\s*[^>]*>)/um)
+						raise ParseException.new( "missing tag start", source, self ) unless md
+						case md[1]
+						when nil
+							return
+						when /\A<[\w_:]/um
+							Element.new(source, self, @context)
+						when /\A<\//um
+							return
+						when CData::START_RE
+							add(CData.new(source))
+						when Comment::START_RE
+							add(Comment.new(source))
+						when Instruction::START_RE
+							add(Instruction.new(source))
+						else
+							add(Text.new(source, whitespace(), raw()))
+						end
+					end
 				end
-				puts "I AM NOW #{self}" if debug
-			rescue ParseException
-				$!.source = source
-				$!.element = self
-				raise
-			rescue Exception
-				old_ex = $!
-				raise ParseException.new("unidentified error", source, self, old_ex)
-			end
-		end
 
-		# A private helper method
-		def parse_children source
-			while true
-				md = source.match(/\A(\s*[^>]*>)/um)
-				#puts "parse_children, match is #{md[0]}, rest is #{source.buffer}"
-				raise ParseException.new( "missing tag start", source, self ) unless md
-				line = md[1]
-				case line
-				when nil
-					return
-				when /\A<[\w_:]/um
-					Element.new(source, self, @context)
-				when /\A<\//um
-					return
-				when CData::START_RE
-					add(CData.new(source))
-				when Comment::START_RE
-					add(Comment.new(source))
-				when Instruction::START_RE
-					add(Instruction.new(source))
-				else
-					add(Text.new(source, whitespace(), raw()))
+				# A private helper method
+				def write_children( writer, indent )
+					cr = indent < 0 ? "" : "\n"
+					#if size == 1 and @children[0].kind_of?(Text)
+					#	self[0].write( writer, -1 )
+					if indent == -1
+						each { |child| child.write( writer, indent ) }
+					else
+						next_indent = indent+2
+						each { |child|
+							writer << cr
+							child.write( writer, next_indent )
+						}
+						writer << cr
+						indent( writer, indent )
+					end
 				end
 			end
-		end
 
-		# A private helper method
-		def write_children( writer, indent )
-			#if size == 1 and @children[0].kind_of?(Text)
-			#	self[0].write( writer, -1 )
-			if indent == -1
-				each { |child| child.write( writer, indent ) }
-			else
-				next_indent = indent+2
-				each { |child|
-					writer << "\n"
-					child.write( writer, next_indent )
-				}
-				writer << "\n"
-				indent( writer, indent )
-			end
-		end
-	end
+			########################################################################
+			# ELEMENTS                                                             #
+			########################################################################
 
-	########################################################################
-	# ELEMENTS                                                             #
-	########################################################################
-
-	# A class which provides filtering of children for Elements, and
-	# XPath search support.
+			# A class which provides filtering of children for Elements, and
+			# XPath search support.
 	class Elements
 		include Enumerable
 		# Constructor
@@ -577,7 +592,8 @@ module REXML
 			if element.kind_of? Element
 				@element.delete element
 			else
-				self[element].remove
+				el = self[element]
+				el.remove if el
 			end
 		end
 
@@ -628,7 +644,7 @@ module REXML
 			@element.each {|child| count+=1 if child.kind_of? Element }
 			count
 		end
-		
+
 		def to_a( xpath=nil )
 			rv = []
 			each( xpath ) { |el| rv << el }
@@ -639,7 +655,7 @@ module REXML
 		# A private helper method
 		def literalize name
 			literal = !(name =~ /^'.*'$/u).nil?
-			name = name[1..-2] if literal
+			name = name[1..-2] if literal and name
 			[literal, name]
 		end
 	end
@@ -725,30 +741,30 @@ module REXML
 				hash = { old_attr.prefix	=> old_attr,
 							value.prefix		=> value }
 							store value.name, hash
-			else
-				store value.name, value
-			end
-			return @element
-		end
+						else
+							store value.name, value
+						end
+						return @element
+					end
 
-		def prefixes
-			ns = []
-			each_attribute do |attribute|
-				ns << attribute.name if attribute.prefix == "xmlns"
-			end
-			ns
-		end
+					def prefixes
+						ns = []
+						each_attribute do |attribute|
+							ns << attribute.name if attribute.prefix == "xmlns"
+						end
+						ns
+					end
 
-		# Removes an attribute
-		# @return the removed attribute
-		def delete( attribute )
-			attr = super
-			attr.remove unless attr.nil?
-			attribute
-		end
+					# Removes an attribute
+					# @return the removed attribute
+					def delete( attribute )
+						attr = super
+						attr.remove unless attr.nil?
+						attribute
+					end
 
-		# Adds an attribute, overriding any existing attribute by the
-		# same name.
+					# Adds an attribute, overriding any existing attribute by the
+					# same name.
 		# @param attribute must be an Attribute
 		def add( attribute )
 			self[attribute.name] = attribute
